@@ -24,9 +24,10 @@ use super::{
 use bytes::Bytes;
 use std::net::SocketAddr;
 
+use backoff::{future::retry, ExponentialBackoff};
 use tokio::sync::broadcast::{self, Receiver, Sender};
 use tokio::sync::mpsc::{self, Receiver as MpscReceiver, Sender as MpscSender};
-use tokio::time::{sleep, timeout, Duration};
+use tokio::time::{timeout, Duration};
 use tracing::{debug, error, info, trace, warn};
 
 /// Host name of the Quic communication certificate used by peers
@@ -43,9 +44,6 @@ const ECHO_SERVICE_QUERY_TIMEOUT: u64 = 30;
 /// Standard size of our channel bounds
 const STANDARD_CHANNEL_SIZE: usize = 10000;
 
-/// Max number of attempts for connection retries
-const MAX_ATTEMPTS: usize = 5;
-
 /// Channel on which incoming messages can be listened to
 pub struct IncomingMessages(pub(crate) MpscReceiver<(SocketAddr, Bytes)>);
 
@@ -302,7 +300,7 @@ impl Endpoint {
         Ok(())
     }
 
-    /// Connects to another peer.
+    /// Connects to another peer, retries for `config.retry_duration_msec` if the connection fails.
     ///
     /// Returns `Connection` which is a handle for sending messages to the peer and
     /// `IncomingMessages` which is a stream of messages received from the peer.
@@ -351,18 +349,7 @@ impl Endpoint {
             Some(Err(error)) => return Err(error.into()),
             None => {}
         }
-
-        let mut attempts: usize = 0;
-
-        let mut connecting = self.attempt_connection(node_addr).await;
-
-        while connecting.is_err() && attempts < MAX_ATTEMPTS {
-            sleep(Duration::from_millis(self.qp2p_config.retry_interval)).await;
-            connecting = self.attempt_connection(node_addr).await;
-            attempts += 1;
-        }
-
-        let final_conn = connecting?;
+        let final_conn = self.attempt_connection(node_addr).await?;
 
         trace!("Successfully connected to peer: {}", node_addr);
 
@@ -375,25 +362,40 @@ impl Endpoint {
         Ok(())
     }
 
-    /// Attempt a connection to a node_addr
+    /// Attempt a connection to a node_addr using exponential backoff
     async fn attempt_connection(&self, node_addr: &SocketAddr) -> Result<quinn::NewConnection> {
-        trace!("Attempting connecton to {:?}", node_addr);
-        let connecting = self.quic_endpoint.connect_with(
-            self.client_cfg.clone(),
-            node_addr,
-            CERT_SERVER_NAME,
-        )?;
+        self.retry(|| async {
+            trace!("Attempting to connect to {:?}", node_addr);
+            let connecting = match self.quic_endpoint.connect_with(
+                self.client_cfg.clone(),
+                node_addr,
+                CERT_SERVER_NAME,
+            ) {
+                Ok(conn) => Ok(conn),
+                Err(error) => {
+                    warn!("Connection attempt failed due to {:?}", error);
+                    Err(Error::from(error))
+                }
+            }?;
 
-        match connecting.await {
-            Ok(new_conn) => Ok(new_conn),
-            Err(error) => {
-                self.connection_deduplicator
-                    .complete(node_addr, Err(error.clone().into()))
-                    .await;
+            let new_conn = match connecting.await {
+                Ok(new_conn) => {
+                    debug!("okay was had");
+                    Ok(new_conn)
+                }
+                Err(error) => {
+                    error!("some error: {:?}", error);
+                    self.connection_deduplicator
+                        .complete(node_addr, Err(error.clone().into()))
+                        .await;
 
-                Err(Error::QuinnConnectionClosed)
-            }
-        }
+                    Err(Error::from(error))
+                }
+            }?;
+
+            Ok(new_conn)
+        })
+        .await
     }
 
     /// Verify if an address is publicly reachable. This will attempt to create
@@ -443,17 +445,7 @@ impl Endpoint {
         &self,
         peer_addr: &SocketAddr,
     ) -> Result<quinn::NewConnection> {
-        let new_connection = match self
-            .quic_endpoint
-            .connect_with(self.client_cfg.clone(), peer_addr, CERT_SERVER_NAME)?
-            .await
-        {
-            Ok(new_connection) => new_connection,
-            Err(error) => {
-                debug!("create new connection check error: {:?}", error);
-                return Err(Error::QuinnConnectionClosed);
-            }
-        };
+        let new_connection = self.attempt_connection(peer_addr).await?;
 
         trace!("Successfully created new connection to peer: {}", peer_addr);
         Ok(new_connection)
@@ -518,17 +510,8 @@ impl Endpoint {
         }
         self.connect_to(dest).await?;
 
-        let mut attempts: usize = 0;
-        let mut res = self.try_send_message(msg.clone(), dest).await;
-
-        while attempts < MAX_ATTEMPTS && res.is_err() {
-            trace!("send attempt # {:?}", attempts);
-            attempts += 1;
-            sleep(Duration::from_millis(self.qp2p_config.retry_interval)).await;
-            res = self.try_send_message(msg.clone(), dest).await;
-        }
-
-        res
+        self.retry(|| async { Ok(self.try_send_message(msg.clone(), dest).await?) })
+            .await
     }
 
     /// Close all the connections of this endpoint immediately and stop accepting new connections.
@@ -579,4 +562,16 @@ impl Endpoint {
     pub fn bootstrap_nodes(&self) -> &[SocketAddr] {
         &self.bootstrap_nodes
     }
+
+    fn retry<I, E, Fn, Fut>(&self, op: Fn) -> impl futures::Future<Output = Result<I, E>>
+    where
+        Fn: FnMut() -> Fut,
+        Fut: futures::Future<Output = Result<I, backoff::Error<E>>>,
+    {
+        let backoff = ExponentialBackoff {
+            max_elapsed_time: Some(Duration::from_millis(self.qp2p_config.retry_duration_msec)),
+            ..Default::default()
+        };
+        retry(backoff, op)
+    }
 }
